# coding: utf-8
from __future__ import unicode_literals, print_function
import json
from pprint import pformat
from mako.template import Template
from io import open
from datetime import datetime, date
from excel2xx.exporter import auto


def _defaultSerialize(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def toJson(excel, output, encoding="utf-8"):
    data = auto.export(excel)
    with open(output, "w", encoding=encoding) as fp:
        json.dump(data, fp, ensure_ascii=False, indent=4, default=_defaultSerialize)
    pass


def toMako(excel, output, template, encoding="utf-8"):
    data = auto.export(excel)
    with open(output, "w", encoding=encoding) as fp:
        text = ""
        with open(template, "r", encoding=encoding) as f:
            text = Template(f.read()).render(excel=data, format=pformat)
        fp.write(text)
    pass


def toMsgPack(excel, output, encoding="utf-8"):
    import msgpack

    data = auto.export(excel)
    with open(output, mode="wb") as fp:
        fp.write(msgpack.packb(data, default=_defaultSerialize))
    pass


def toCSV(excel, output, encoding="utf-8"):
    # data = auto.export(excel)
    raise NotImplementedError("toCSV")
    pass



from colorama import Fore


class Colors:
    @staticmethod
    def result(text, is_ok=True):
        if is_ok:
            text = Fore.LIGHTGREEN_EX + str(text) + Fore.RESET
            return " [" + Colors.green(text) + "]"
        else:
            text = Fore.LIGHTRED_EX + str(text) + Fore.RESET
            return " [" + Colors.red(text) + "]"
        pass

    @staticmethod
    def green(text):
        return Fore.LIGHTGREEN_EX + str(text) + Fore.RESET

    @staticmethod
    def red(text):
        return Fore.LIGHTRED_EX + str(text) + Fore.RESET

    @staticmethod
    def blue(text):
        return Fore.LIGHTBLUE_EX + str(text) + Fore.RESET

    @staticmethod
    def yellow(text):
        return Fore.LIGHTYELLOW_EX + str(text) + Fore.RESET

    @staticmethod
    def magenta(text):
        return Fore.LIGHTMAGENTA_EX + str(text) + Fore.RESET



import os
import excel2xx.codegen as codegen
import logging
import tempfile

root_dir = os.path.dirname(os.path.dirname(__file__))
test_dir = os.path.dirname(__file__)

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def test_codegen_md():
    with tempfile.TemporaryDirectory(prefix="codegen-") as d:
        log.info("temp directory: %s", d)
        for fname in ["abc", "def", "ghi"]:
            write_file(d + f"/{fname}.md.mako", example_md[0])
            pass

        ctx = {
            "from": "testcase",
        }
        src = d + "/*.mako"
        codegen.generate(src, ctx, note="")
        for fname in ["abc", "def", "ghi"]:
            expected = example_md[1].replace("${__file__}", d + f"/{fname}.md.mako")
            assert expected == read_file(d + f"/{fname}.md")
        pass
    pass


def read_file(fpath) -> str:
    log.info("read_file %s", fpath)
    with open(fpath) as fp:
        return fp.read()
    pass


def write_file(fpath, content: str):
    log.info("write_file %s", fpath)
    with open(fpath, "w") as fp:
        return fp.write(content)
    pass


example_md = [
    """
This is generated by template file "${__file__}"

```
% for i in range(3):
${i} ** 2 = ${i**2}
% endfor
```
""",
    """
This is generated by template file "${__file__}"

```
0 ** 2 = 0
1 ** 2 = 1
2 ** 2 = 4
```
""",
]



import os
from excel2xx import Excel, fields, fieldmeta

root_dir = os.path.dirname(os.path.dirname(__file__))


def test_fields():
    fpath = root_dir + "/example/test.xlsx"
    rows = Excel(fpath)[0].toList()
    assert rows
    assert rows[0]["fieldJson"] == {"a": 1, "b": 2}
    assert rows[0]["fieldJson"] != {"a": 1, "b": 2, "c": 3}
    d = Excel(fpath)[0].toDict()
    assert d


def test_IntArray():
    f = fieldmeta.DEFINE_FIELDS["array<int>"]
    assert f == fields.IntArray
    assert f.format(None, 1.0) == [1]
    assert f.format(None, "1,2, 6,   7,8") == [1, 2, 6, 7, 8]
    assert f.format(None, "1, 9 ,2,3") == [1, 9, 2, 3]



# encoding: utf-8
import os
import sys

debug = os.environ.get("EXCEL2XX_DEBUG", 0)
if debug:
    print("sys.path:")
    for path in sys.path:
        print("\t" + path)

import excel2xx.main
import excel2xx.fields

if __name__ == "__main__":
    excel2xx.setUnits("K,M,G,T,P", size=1024)
    exit(excel2xx.main.main_docopt())
    pass



# encoding: utf-8
import os
import importlib
import importlib.util
from excel2xx import console


def import_validator(name, fpath):
    spec = importlib.util.spec_from_file_location(name, fpath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # sys.modules[name] = module
    return module


def run(data, data_dir: str, validator_name="_validator"):
    validator_dir = os.path.join(data_dir, validator_name)
    for fpath in list_files(validator_dir):
        show = fpath[len(data_dir) :].strip("/")
        # fpathx = console.Colors.blue(show)
        print(f"\t{show} -> ", end="")
        # mpath = fpath.replace("/", ".").replace(".py", "")
        pkgname = os.path.basename(fpath).split(".")[0]
        mod = import_validator(pkgname, fpath)
        try:
            mod.validate(data)
            console.success("OK")
        except Exception as ex:
            console.fail("FALIED")
            console.error("\t\t" + str(ex))
            return 1
        pass
    print("\n")
    return 0


def list_files(_dir: str):
    import bisect

    rs = []  # type: list[str]
    for base, dirs, files in os.walk(_dir):
        for fname in files:
            if not fname.endswith(".py"):
                continue
            if fname.endswith("__.py"):
                continue
            fpath = os.path.join(base, fname)
            # rs.append(fpath)
            bisect.insort(rs, fpath)
            pass
        pass
    return rs



# encoding: utf-8
from __future__ import unicode_literals, print_function
import os
import xlrd
from collections import OrderedDict
from excel2xx import utils, fields, fieldmeta, console
import excel2xx.exporter as exporter
from .version import VERSION

_ = VERSION

__author__ = "cupen"
__email__ = "xcupen@gmail.com"

FieldMeta = fieldmeta.FieldMeta


class Excel:
    def __init__(self, fpathOrFp, fieldMeta=FieldMeta(name=0, type=1, desc=2, data=3)):
        self.__filePath = fpathOrFp
        self.fieldMeta = fieldMeta

        self.__callback = None
        if isinstance(self.__filePath, str):
            self.__wb = xlrd.open_workbook(self.__filePath)
        else:
            self.__wb = xlrd.open_workbook(file_contents=self.__filePath.read())
        pass

    @property
    def datemode(self):
        return self.__wb.datemode

    @property
    def fname(self):
        if isinstance(self.__filePath, str):
            return os.path.basename(self.__filePath)

        with open("/dev/null") as f:
            fileType = type(f)
            if isinstance(self.__filePath, fileType):
                return self.__filePath.name

        return ""

    def getSheet(self, sheetName, alias=""):
        sheet = None
        try:
            sheet = self.__wb.sheet_by_name(sheetName)
        except xlrd.XLRDError as e:
            if not alias:
                raise e
            sheet = self.__wb.sheet_by_name(alias)
        return Sheet(self, sheet)

    def __getitem__(self, nameOrIdx):
        if isinstance(nameOrIdx, int):
            sheet = self.__wb.sheet_by_index(nameOrIdx)
            return Sheet(self, sheet)
        return self.getSheet(nameOrIdx)

    def __iter__(self):
        sheets = map(lambda x: Sheet(self, x), self.__wb.sheets())
        for sheet in sheets:
            if sheet.name.startswith("#"):
                continue
            yield sheet
        pass

    def toList(self):
        _dict = OrderedDict()
        for sheet in self:
            _dict[sheet.name] = sheet.toList()
        return _dict

    def toDict(self):
        _dict = OrderedDict()
        for sheet in self:
            _dict[sheet.name] = sheet.toDict()
        return _dict


class Sheet:
    def __init__(self, excel, sheet):
        self.__excel = excel
        self.__sheet = sheet
        self.__fields = {}
        pass

    @property
    def name(self):
        return self.__sheet.name

    def fields(self):
        """
        :rtype: dict of [str, Field]
        """
        sheet = self.__sheet
        if len(self.__fields) <= 0:
            self.__fields = self.__excel.fieldMeta.parseSheet(self.__excel, sheet)
        return self.__fields.copy()

    def rows(self):
        skipRows = self.__excel.fieldMeta.dataRowIdx
        emptyRows = []
        for row in self.__sheet.get_rows():
            if skipRows > 0:
                skipRows -= 1
                # print(row)
                continue

            if len(row) <= 0:
                emptyRows += 1
                continue

            if self._isEmptyRow(row):
                emptyRows.append(row)
                continue
            elif len(emptyRows) > 0:
                raise self.throwException(f"Invalid row with empty'{emptyRows[0]}'")
            yield row
        pass

    def _isEmptyRow(self, row):
        _fields = self.fields()
        len(_fields)
        if len(row) > len(_fields):
            row = row[: len(_fields)]
        for cell in row:
            if cell.ctype != xlrd.XL_CELL_EMPTY:
                return False
            pass
        return True

    def toList(self):
        return list(iter(self))

    def firstFieldName(self):
        fields = self.fields()
        for fieldName in fields:
            return fieldName
        return None

    def throwException(self, text):
        raise Exception(f"Sheet(name={self.name}): {text}")

    def toDict(self, valueIsList=False):
        _dict = OrderedDict()
        firstField = self.firstFieldName()
        if not firstField:
            return self.throwException(f'Invalid first field name. "{firstField}"')
        for row in self:
            if not row:
                continue

            value = row[firstField]
            if (not valueIsList) and (value in _dict):
                return self.throwException(
                    "Duplicate value of field name. %s=%s" % (firstField, value)
                )
            _dict[value] = row
        return _dict

    def toDict2(self):
        _dict = OrderedDict()
        firstField = self.firstFieldName()
        if not firstField:
            return self.throwException("Invalid first field name. %s" % firstField)
        for row in self:
            if not row:
                continue

            value = row[firstField]
            if value not in _dict:
                _dict[value] = []
            _dict[value].append(row)
        return _dict

    def toKV(self):
        return exporter.toKV(self)

    def toDataFrame(self):
        import pandas

        return pandas.DataFrame(self, columns=self.fields().keys())

    @property
    def fname(self):
        return self.__excel.fname

    def __iter__(self):
        fields = self.fields()
        fieldsArr = tuple(fields.values())
        rowNum = 0
        for row in self.rows():
            rowNum += 1
            row = list(row)

            _dict = OrderedDict()
            for i in range(len(fieldsArr)):
                cell = row[i]
                field = fieldsArr[i]
                try:
                    _dict[field.name] = field.format(cell.value)
                except Exception as ex:
                    console.error(
                        "%-12s : Field(name=%s type=%s). row: %s col: %s\n\t err: %s\n\t %s\n"
                        % (
                            self.fname,
                            field.name,
                            field.type,
                            utils.show_row(rowNum),
                            utils.show_col(i),
                            str(ex),
                            cell.value,
                        )
                    )
                    pass

            yield _dict
        pass


def addFieldType(typeName, parser):
    fieldmeta.DEFINE_FIELDS[typeName] = parser
    pass


def delFieldType(typeName, parser):
    del fieldmeta.DEFINE_FIELDS[typeName]
    pass


def setUnits(text, size=1000):
    fields.BigNumber.setUnits(text, size=size)
    fields.ItemExpr.setUnits(text)
    pass



import os
import json
from conftest import testdata_dir
import openpyxl
# import pytest

excel_dir = os.path.join(testdata_dir, "_excel")
excel_files = [
    os.path.join(excel_dir, "test-default.xlsx"),
    os.path.join(excel_dir, "test-data-row.xlsx"),
]


# @pytest.fixture(scope="module")
def setup_module():
    os.makedirs(excel_dir, exist_ok=True)
    pass


def F(fname):
    return os.path.join(excel_dir, fname)


def test_default():
    gen_default()
    from excel2xx import main

    argv = ["json", "-v", "-o", F("yes.json"), excel_files[0]]
    code = main.main_docopt(argv)
    assert 0 == code
    with open(F("yes.json")) as fp:
        data = json.load(fp)
        pass
    assert 3 == len(data["auto"])
    assert 1 == len(data["sheet0"])
    assert 1 == len(data["sheet1"])
    assert 1 == len(data["sheet2"])
    assert 3 == len(data["sheet-map"])
    pass


def test_datarow():
    gen_datarow()
    from excel2xx import main

    argv = ["json", "--data-row", "6", "-o", F("yes.json"), excel_files[1]]
    code = main.main_docopt(argv)
    assert 0 == code
    with open(F("yes.json")) as fp:
        data = json.load(fp)
        pass
    assert 3 == len(data["auto"])
    assert 1 == len(data["sheet0"])
    assert 1 == len(data["sheet1"])
    assert 1 == len(data["sheet2"])
    assert 3 == len(data["sheet-map"])
    pass


def gen_default():
    setup_module()
    w = openpyxl.Workbook()
    s = w.create_sheet(title="auto")
    s.append(["key", "type", "value"])
    s.append(["string", "auto", "auto"])
    s.append(["", "", ""])
    s.append(["key1", "int", 1])
    s.append(["key2", "string", "2"])
    s.append(["key3", "array<int>", "1,2,3"])
    for i in range(3):
        s = w.create_sheet(title=f"sheet{i}")
        s.append(["key1", "key2", "key3", "key3"])
        s.append(["int", "string", "float", "array<int>"])
        s.append(["", "", "", ""])
        s.append([1, "2", 3.0, "1,2,3"])
        pass

    s = w.create_sheet("sheet-map(map)")
    s.append(["key1", "key2"])
    s.append(["int", "string"])
    s.append(["", ""])
    s.append([1, "1"])
    s.append([2, "2"])
    s.append([3, "3"])

    w.save(excel_files[0])
    pass


def gen_datarow():
    setup_module()
    w = openpyxl.Workbook()
    s = w.create_sheet(title="auto")
    s.append(["key", "type", "value"])
    s.append(["string", "auto", "auto"])
    s.append(["", "", ""])
    s.append(["", "", ""])
    s.append(["", "", ""])
    s.append(["key1", "int", 1])
    s.append(["key2", "string", "2"])
    s.append(["key3", "array<int>", "1,2,3"])
    for i in range(3):
        s = w.create_sheet(title=f"sheet{i}")
        s.append(["key1", "key2", "key3", "key3"])
        s.append(["int", "string", "float", "array<int>"])
        s.append(["", "", "", ""])
        s.append(["", "", "", ""])
        s.append(["", "", "", ""])

        s.append([1, "2", 3.0, "1,2,3"])
        pass

    s = w.create_sheet("sheet-map(map)")
    s.append(["key1", "key2"])
    s.append(["int", "string"])
    s.append(["", ""])
    s.append(["", ""])
    s.append(["", ""])
    s.append([1, "1"])
    s.append([2, "2"])
    s.append([3, "3"])

    w.save(excel_files[1])
    pass



# encoding: utf-8


def export(ex, ver=1):
    d = {}
    for sheet in ex:
        if str.startswith(sheet.name, "#"):
            continue
        name, _type = parse_sheet_name(sheet.name)
        if not name:
            raise Exception(f"invalid sheet {sheet.name}")
        if _type == "list":
            d[name] = sheet.toList()
        elif _type == "map":
            d[name] = sheet.toDict()
        elif _type == "kv":
            d[name] = sheet.toKV()
        else:
            raise Exception(f"invalid sheet {sheet.name} name={name} type={_type}")
    return d


def parse_sheet_name(name: str) -> (str, str):
    """
    >>> _parse_sheet_name("name")
    ('name', 'list')
    >>> _parse_sheet_name("name(list)")
    ('name', 'list')
    >>> _parse_sheet_name("name123(map)")
    ('name123', 'map')
    >>> _parse_sheet_name("name123((map)")
    Traceback (most recent call last):
    Exception: duplicated symbol '('
    >>> _parse_sheet_name("sheet-map(map)")
    ('sheet-map', 'map')
    """
    name = name.strip()
    if "=" in name:
        arr = list(map(lambda x: x.strip(), name.split("=")))
        name = arr[1]
    return _parse_sheet_name(name)


def _parse_sheet_name(name) -> (str, str):
    realName = ""
    realType = ""
    state = 0  # 0-start, 1-name, 2-type, 3-end
    for c in name:
        if c == "(":
            if state != 1:
                raise Exception("duplicated symbol '('")
            state = 2
        elif c == ")":
            if state != 2:
                raise Exception(" ')' must behind of '('")
            if len(realType) <= 0:
                raise Exception(" '()' must be one of list, map, kv")
            state = 3
            break  # end
        else:
            if not str.isalnum(c) and c not in ["-", "_"]:
                raise Exception(f"invalid char '{c}'")
            if state == 0:
                realName = c
                state = 1
            elif state == 1:
                realName += c
            elif state == 2:
                realType += c
            else:
                raise Exception("BUG!")
            pass
        pass
    if realType == "":
        realType = "list"
    return realName, realType



# encoding: utf-8
from typing import Optional
from collections import OrderedDict
import excel2xx.fields as fields

DEFINE_FIELDS = {
    "": fields.Auto,
    "auto": fields.Auto,
    "bool": fields.Bool,
    "int": fields.Int,
    "number": fields.Int,
    "float": fields.Float,
    "str": fields.String,
    "string": fields.String,
    "array": fields.Array,
    "array<int>": fields.IntArray,
    "array<float>": fields.FloatArray,
    "array<string>": fields.Array,
    "map<string,string>": fields.Map,
    "map<string, string>": fields.Map,
    "date": fields.Date,
    "datetime": fields.DateTime,
    "unixstamp": fields.UnixStamp,
    "object": fields.Object,
    "array<object>": fields.ObjectArray,
    "itemexpr": fields.ItemExpr,
    "array<itemexpr>": fields.ItemExprArray,
    "ItemExpr": fields.ItemExpr,
    "array<ItemExpr>": fields.ItemExprArray,
    "Reward": fields.Reward,
    "reward": fields.Reward,
    "Ratio": fields.Ratio,
    "ratio": fields.Ratio,
    "bignumber": fields.BigNumber,
    "BigNumber": fields.BigNumber,
    "json": fields.JSON,
}


class FieldMeta:
    def __init__(self, name=0, type=1, desc=2, data=3):
        self._lineNumCfg = {
            "name": name,
            "type": type,
            "desc": desc,
            "data": data,
        }
        pass

    @property
    def nameRowIdx(self):
        return self._lineNumCfg["name"]

    @property
    def typeRowIdx(self):
        return self._lineNumCfg["type"]

    @property
    def descRowIdx(self):
        return self._lineNumCfg["desc"]

    @property
    def dataRowIdx(self):
        return self._lineNumCfg["data"]

    @classmethod
    def parseFieldType(cls, fieldType: str) -> str:
        fieldType = fieldType.replace(" ", "")
        if fieldType.startswith("object") or fieldType.startswith("Object"):
            return "object"

        if fieldType.startswith("array<object") or fieldType.startswith("array<Object"):
            return "array<object>"
        return fieldType

    @classmethod
    def parseField(cls, name: str, fieldType: str) -> Optional[fields.Field]:
        name = cls.parseFieldType(fieldType)
        meta = DEFINE_FIELDS.get(name)
        if not meta:
            return None
        return meta(name, fieldType)

    def parseSheet(self, excel, sheet):
        if sheet.nrows < self.nameRowIdx + 1:
            return {}

        nameRow = sheet.row(self.nameRowIdx)
        typeRow = sheet.row(self.typeRowIdx)
        # descRow = sheet.row(self.descRowIdx)

        fields = OrderedDict()
        for i in range(0, len(nameRow)):
            fieldName = str(nameRow[i].value).strip()
            fieldType = str(typeRow[i].value).strip()

            if not fieldName or fieldName.startswith("#"):
                continue
            fieldMeta = DEFINE_FIELDS.get(self.parseFieldType(fieldType))
            if not fieldMeta:
                raise RuntimeError(
                    '%-12s : Unexist field meta "%s". check the field(%s)'
                    % (excel.fname, fieldType, repr(typeRow[i]))
                )

            fields[fieldName] = fieldMeta(fieldName, fieldType, excel)
            i += 1
            pass
        return fields

    pass



# encoding: utf-8
import os
import logging
import glob
import traceback
import excel2xx.console as console
from mako.template import Template

log = logging.getLogger("codegen")

NOTE = "// DONT EDIT: generated by excel2xx/codegen.\n"


def generate(src: list, context={}, note=NOTE, encoding="utf-8", newline="\n"):
    """
    generate source code from mako template files
    """
    if isinstance(src, str):
        src = [src]
    elif isinstance(src, list):
        pass
    else:
        raise Exception("invalid params: src = " + str(src))

    fpaths = []
    for pattern in src:
        fpaths += glob.glob(pattern, recursive=True)
        pass
    print(f"\ttemplate files: {len(fpaths)}")

    for fpath in fpaths:
        fpath_new = fpath.replace(".mako", "")
        print(f"\t{fpath} -> ", end="")
        if not fpath.endswith(".mako"):
            print(console.Colors.yellow(fpath_new))
            continue

        ctx = dict(**context)
        ctx["__file__"] = fpath
        ctx["__dir__"] = os.path.dirname(fpath)

        tmpl = load_template(fpath)
        try:
            text = tmpl.render(**ctx)
            with open(fpath_new, "w", encoding=encoding, newline=newline) as fp:
                if note:
                    fp.write(note)
                fp.write(text)
                pass
            print(console.Colors.green(fpath_new))
            if fpath_new.endswith(".go"):
                os.system(f"go fmt {fpath_new} 2>&1 > /dev/null")
        except Exception:
            print(console.Colors.red(fpath_new))
            print("")
            print(console.Colors.red(f"something wrong with '{fpath}'"))
            stacks = traceback.format_exc()
            print(stacks)
            exit(101)
            pass
        pass
    print("\n")
    pass


def load_template(fpath) -> Template:
    with open(fpath, mode="r", encoding="utf-8") as fp:
        return Template(fp.read())
    pass



# encoding: utf-8
from __future__ import unicode_literals, print_function

import datetime
import re
import json
import math
from collections import namedtuple, OrderedDict
from pprint import pformat

from xlrd.xldate import xldate_as_datetime


class Field:
    def __init__(self, name, type, wb=None):
        self.name = name
        self.type = type
        self.wb = wb
        pass

    def format(self, v):
        raise NotImplementedError

    @classmethod
    def as_type(cls, typeName):
        typeName = typeName.strip()
        if typeName == "int":
            return int
        if typeName == "string":
            return str
        if typeName == "float":
            return float
        if typeName == "ItemExpr":
            return ItemExpr("as_type", typeName).format
        raise Exception("Invalid type %s" % typeName)
        pass

    pass


class Int(Field):
    def format(self, v):
        return int(v)


class Bool(Field):
    TRUE_VALUES = [1, 1.0, "true"]
    FALSE_VALUES = [0, 0.0, "false"]

    def format(self, v):
        # print(f"{type(v)}: {v}", end="")
        if v in self.TRUE_VALUES:
            return True
        if v in self.FALSE_VALUES:
            return False
        raise Exception(f"Invalid BoolValue: {type(v)}:({v})")
        pass


class String(Field):
    def format(self, v):
        v = str(v)
        v = v.replace("\\r\\n", "\n")
        v = v.replace("\\n", "\n")
        v = v.replace("\\r", "\n")
        return v


class Float(Field):
    """
    recommend using string, beacuse excel-float was'nt adhere IEEE 754.
    @see https://support.microsoft.com/en-us/help/78113/floating-point-arithmetic-may-give-inaccurate-results-in-excel
    """

    def format(self, v):
        return float(v)


class Array(Field):
    def format(self, v):
        if not v:
            return []
        _iter = map(lambda x: x.strip(), v.split(","))
        return list(_iter)


class IntArray(Field):
    def format(self, v):
        if not v:
            return []
        if isinstance(v, float):
            if v != int(v):
                raise Exception("Invalid cell type float")
            return [int(v)]
        elif isinstance(v, int):
            return [v]
        elif isinstance(v, str):
            _iter = map(lambda x: x.strip(), v.split(","))
            return list(map(int, _iter))
        else:
            raise Exception("Invalid cell type " + str(type(v)))
        return


class Ratio(Field):
    def format(self, v):
        try:
            if not v:
                raise Exception("empty value")
            arr = v.split(":")
            if len(arr) != 2:
                raise Exception("it must be two numbers.")
            _iter = map(lambda x: x.strip(), arr)
            return list(map(int, _iter))
        except Exception as ex:
            raise Exception(f"invalid Ratio: {v}. err:{ex}")


class FloatArray(Field):
    def format(self, v):
        if not v:
            return []
        _iter = map(lambda x: x.strip(), v.split(","))
        return list(map(float, _iter))


class Auto(Field):
    def format(self, v):
        return v


class Date2(Field):
    FORMAT = "%Y-%m-%d"

    def format(self, v):
        return datetime.datetime.strptime(v, self.FORMAT)


class DateTime2(Field):
    FORMAT = "%Y-%m-%dT%H:%M:%S%z"

    def format(self, v):
        return datetime.datetime.strptime(v, self.FORMAT)


class UnixStamp(DateTime2):
    def format(self, v):
        dt = super().format(v)
        return int(dt.timestamp() * 1000)


class Date(Field):
    def format(self, v):
        try:
            return xldate_as_datetime(v, self.wb.datemode).date()
        except Exception:
            return None


class DateTime(Field):
    def format(self, v):
        try:
            return xldate_as_datetime(v, self.wb.datemode)
        except Exception:
            return None


class Map(Field):
    def format(self, v: str) -> dict:
        rs = {}
        m = filter(lambda x: bool(x), v.split("\n"))
        for line in m:
            arr = line.split(":", maxsplit=1)
            if len(arr) != 2:
                raise Exception(f"Invalid map item {pformat(line)}")
            arr = list(map(lambda x: x.strip(), arr))
            rs[arr[0]] = arr[1]
        return rs


class Object(Field):
    Attr = namedtuple("Attr", ["name", "type"])

    def __init__(self, name, type, wb=None):
        super(Object, self).__init__(name, type, wb)
        self.attrs = self.parseType(text=type)
        pass

    def newException(self):
        return Exception(
            "Invalid object define. name:%s type:%s attrs:%s"
            % (self.name, self.type, self.attrs)
        )

    def parseType(self, text):
        text = text.strip()
        if not text.startswith("object"):
            raise self.newException()

        attrDefs = (
            text.replace("Object", "")
            .replace("object", "")
            .replace(" ", "")
            .strip("()")
            .split(",")
        )
        if len(attrDefs) <= 0:
            raise self.newException()

        attrs = []
        for attrDefine in attrDefs:
            tmpArr = attrDefine.split(":")
            if len(tmpArr) == 1:
                name = tmpArr[0].strip()
                type = str
            elif len(tmpArr) >= 2:
                name = tmpArr[0].strip()
                type = self.as_type(tmpArr[1])
            else:
                raise Exception(
                    'Warning: Invalid fieldType="%s" tmpArr=%s' % (text, tmpArr)
                )
                # print("Warning: Invalid fieldType %s" % text)
                name = tmpArr[0].strip()
                type = str

            attrs.append(self.Attr(name=name, type=type))
            pass
        return attrs

    def parseValue(self, attrs, valText):
        if not valText:
            return None
        vals = list(map(lambda x: x.strip(), valText.strip("{}<> ").split(",")))
        if len(vals) < len(self.attrs) - 1:
            attrs = ",".join(map(str, self.attrs))
            raise Exception(
                "Invalid object define. name:%s type:%s attrs:[%s] val:%s"
                % (self.name, self.type, attrs, valText)
            )

        d = OrderedDict()
        for i in range(len(vals)):
            attr = attrs[i]
            # if i >= len(vals):
            #     d[attr.name] = attr.type()
            #     continue
            d[attr.name] = attr.type(vals[i])
            pass
        return d

    def format(self, v):
        return self.parseValue(self.attrs, v)


class ObjectArray(Object):
    pattern = re.compile("{[^{^}]*}")

    def __init__(self, name, type, wb=None):
        _type = type.replace("array", "").strip("<>")
        super(ObjectArray, self).__init__(name, _type, wb=wb)
        pass

    def newException(self):
        return Exception(
            "Invalid array<object> define. name:%s type:%s attrs:%s"
            % (self.name, self.type, self.attrs)
        )

    def format(self, v):
        if not v:
            return []

        _list = []
        for tmpVal in self.pattern.findall(v):
            obj = super(ObjectArray, self).format(tmpVal.strip())
            _list.append(obj)
        return _list


class ItemExpr(Field):
    parseFuncs = {}

    @staticmethod
    def Default(arr):
        if len(arr) < 2 or len(arr) > 3:
            raise Exception(f"invalid itemexpr. {arr}")
        return {
            "type": arr[0],
            "id": arr[1],
            "count": int(arr[2]) if len(arr) >= 3 else 1,
        }

    @staticmethod
    def OnlyID(arr):
        if len(arr) < 2:
            raise Exception(f"invalid itemexpr-OnlyID. {arr}")
        if len(arr) >= 3 and arr[2] != "1":
            raise Exception(f"invalid itemexpr-OnlyID. {arr}")
        return {"type": arr[0], "id": arr[1], "count": 1}

    @staticmethod
    def OnlyCount(arr):
        if len(arr) != 2:
            raise Exception(f"invalid itemexpr-OnlyCount. {arr}")
        return {"type": arr[0], "count": int(arr[1])}

    @staticmethod
    def FloatCount(arr):
        if len(arr) != 2:
            raise Exception(f"invalid itemexpr-FloatCount. {arr}")
        return {"type": arr[0], "count": Number("", "").format(arr[1])}

    def __init__(self, name, type, wb=None):
        super(ItemExpr, self).__init__(name, type, wb)
        pass

    @classmethod
    def register(cls, name, parseFunc):
        if name in cls.parseFuncs:
            raise Exception(f"already registered. name:{name} parseFunc:{parseFunc}")
        cls.parseFuncs[name] = parseFunc
        pass

    @classmethod
    def setUnits(cls, units: str):
        units = units.strip(" \n\r,")
        units = "," + units

        idx = 0
        _dict = {}
        for v in map(lambda x: x.strip(), units.split(",")):
            _dict[v] = idx
            idx += 1
            pass
        cls.UNITS = _dict
        pass

    def newException(self, value):
        return Exception(
            "Invalid ItemExpr. name:%s type:%s value:%s" % (self.name, self.type, value)
        )

    def format(self, v):
        if not isinstance(v, (str, list)):
            raise self.newException(v)

        if v == "":
            return None

        tmpArr = v
        if isinstance(tmpArr, str):
            tmpArr = v.split("-")

        if len(tmpArr) <= 0:
            raise self.newException(v)

        _type = tmpArr[0]
        parseFunc = self.parseFuncs.get(_type, ItemExpr.Default)
        return parseFunc(tmpArr)


class ItemExprArray(ItemExpr):
    def newException(self, value):
        return Exception(
            "Invalid Array<ItemExpr>. name:%s type:%s value:%s"
            % (self.name, self.type, value)
        )

    def format(self, v):
        if not isinstance(v, str):
            raise self.newException(v)

        if not v:
            return []
        _iter = map(lambda x: x.strip(), v.split(","))
        _iter = map(lambda x: super(ItemExprArray, self).format(x), _iter)
        return list(_iter)


class Reward(ItemExpr):
    def newException(self, value):
        return Exception(
            "Invalid Array<ItemExpr>. name:%s type:%s value:%s"
            % (self.name, self.type, value)
        )

    def parseItemExprWithWeight(self, expr_w):
        tmpArr = expr_w.split(",")
        weight = int(tmpArr[0])
        if len(tmpArr) < 2:
            raise self.newException(expr_w)

        exprObj = ItemExpr.format(self, tmpArr[1])
        return {
            "weight": weight,
            "item": exprObj,
        }

    def format(self, v):
        if not v:
            return []
        if not isinstance(v, str):
            raise self.newException(v)

        _iter = map(lambda x: x.strip(), v.split(";"))
        _iter = map(lambda x: self.parseItemExprWithWeight(x), _iter)
        return list(_iter)


class Number(Field):
    """
    Float64

    >>> n = Number("test", type="number")
    >>> n.format("1K")
    1024.0
    >>> n.format("1.2K")
    1228.8
    >>> n.format("512.1048576M")
    536980863.1627775
    """

    UNITS = {}
    UNITS_IDX = {}
    PARTERN = re.compile(r"(?P<count>\d+\.?\d*)(?P<unit>[a-zA-Z_]*)")

    def newException(self, value, suffix=""):
        return Exception(
            "Invalid BigNumber. name:%s type:%s value:%s. suffix:%s"
            % (self.name, self.type, value, suffix)
        )

    @classmethod
    def setUnits(cls, units, size=1000):
        units = units.strip(" \n\r,")
        if isinstance(units, str):
            units = map(lambda x: x.strip(), units.split(","))

        idx = 0
        _dict = {}
        _dict2 = {}
        for v in units:
            idx += 1
            _dict[v] = idx - 1
            _dict2[v] = math.pow(size, idx)
            pass

        cls.UNITS_IDX = _dict
        cls.UNITS = _dict2
        pass

    def format(self, v):
        if isinstance(v, float):
            return v

        if not v or not isinstance(v, str):
            raise self.newException(v)

        _dict = self.PARTERN.match(v).groupdict()
        if "count" not in _dict:
            raise self.newException(v)
        if "unit" not in _dict:
            raise self.newException(v)
        if _dict["unit"] and _dict["unit"] not in self.UNITS:
            raise self.newException(v, suffix=f"Invalid UNITS {self.UNITS}")
        unit = 1
        if _dict["unit"] != "":
            unit = self.UNITS[_dict["unit"]]
        count = float(_dict["count"])
        return count * unit


Number.setUnits("K,M,G,T,P,E", size=1024)
BigNumber = Number


class JSON(ItemExpr):
    def newException(self, value):
        return Exception(
            "invalid json. name:%s type:%s value:%s" % (self.name, self.type, value)
        )

    def parseJson(self, value):
        return json.loads(value)

    def format(self, v):
        if not v:
            return None
        if not isinstance(v, str):
            raise self.newException(v)
        return self.parseJson(v)


class Duration(object):
    def newException(self, value):
        return Exception(
            "invalid duration. name:%s type:%s value:%s" % (self.name, self.type, value)
        )

    def parseDuration(self, value):
        if not isinstance(value, str):
            raise self.newException(value)
        state = 0
        for c in value:
            pass
        return 0

    def format(self, v):
        raise NotImplementedError("Duration.format")
        # return self.parseDuration(v)



# coding: utf-8
import colorama
from colorama import Fore
from .colors import Colors
__all__ = ["Colors"]

# from .colors import Colors
colorama.init(autoreset=True)


def info(text):
    print(Fore.LIGHTCYAN_EX + str(text) + Fore.RESET)


def warn(text):
    print(Fore.LIGHTYELLOW_EX + str(text) + Fore.RESET)


def error(text):
    print(Fore.LIGHTRED_EX + str(text) + Fore.RESET)


def success(text):
    text = Fore.LIGHTGREEN_EX + str(text) + Fore.RESET
    print(f"[{text}]")


def fail(text):
    text = Fore.LIGHTRED_EX + str(text) + Fore.RESET
    print(f"[{text}]")



# encoding: utf-8
"""Export data from Excel file.

Usage:
    excel2xx.py json    <excel> [options]
    excel2xx.py msgpack <excel> [options]
    excel2xx.py mako    <excel> --template=FILE [options]
    excel2xx.py --version

Arguments:
    <excel>                        Excel file path.

Options:
    -h --help                      show this help message and exit.
    --version                      show version and exit.
    -o --output=FILE               output to file.
    --name-row=NAME_ROW            name row number. [default: 1]
    --type-row=TYPE_ROW            type row number. [default: 2]
    --desc-row=DESC_ROW            desc row number. [default: 3]
    --data-row=DATA_ROW            data row number. [default: 4]
    -v --verbose                   show debug infomation.
    -vv --verbose2                 show more debug infomation.
"""

import os
import traceback
from docopt import docopt
from excel2xx import Excel, export, FieldMeta
from excel2xx.version import VERSION


def main(args):
    if args["--version"]:
        print(VERSION)
        return 0
    src = args["<excel>"]
    dest = args["--output"]

    if src:
        src = os.path.realpath(src)
    if dest:
        dest = os.path.realpath(dest)

    if not os.path.isfile(src):
        print("Unexist file:" + src)
        return 1

    meta = FieldMeta(
        name=int(args["--name-row"]) - 1,
        type=int(args["--type-row"]) - 1,
        desc=int(args["--desc-row"]) - 1,
        data=int(args["--data-row"]) - 1,
    )
    excel = Excel(src, fieldMeta=meta)
    if args["json"]:
        dest = dest or f"{src}.json"
        export.toJson(excel, dest)
    elif args["msgpack"]:
        dest = dest or f"{src}.msgp"
        export.toMsgPack(excel, dest)
    elif args["mako"]:
        template = args["--template"]
        if not dest:
            print("missing '--template', which is mako template file.")
            return 1
        export.toMako(excel, dest, template)
    else:
        print("Invalid subcmd.")
        return 2
    return 0


def main_docopt(argv=None):
    args = docopt(__doc__, argv)
    if args["--verbose2"]:
        print(args)
    try:
        return main(args)
    except Exception as e:
        errorMsg = traceback.format_exc() if args["--verbose"] else e
        print(errorMsg)
        return 2
    pass


if __name__ == "__main__":
    import sys

    exit(main_docopt(sys.argv))



from setuptools import setup, find_packages
from io import open
import os

root_dir = os.path.dirname(__file__)


def _open(fname):
    return open(os.path.join(root_dir, fname), "r", encoding="utf-8")


with _open("README.md") as fp:
    readme = fp.read()
    pass

with _open("requirements.txt") as fp:
    deps = map(str.strip, fp.readlines())
    deps = filter(lambda line: bool(line), deps)
    deps = list(deps)
    pass

print(deps)


setup(
    name="excel2xx",
    version="0.11.6",
    packages=find_packages(),
    url="https://github.com/cupen/excel2xx",
    license="WTFPL",
    author="cupen",
    author_email="xcupen@gmail.com",
    description="Extract data from excel file, and export to json, msgpack, or any code(mako template).",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=deps,
    entry_points={
        "console_scripts": [
            "excel2xx=excel2xx.main:main_docopt",
        ],
    },
    # python_requires=">=3.6,!=3.6.1",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# encoding: utf-8
from collections import OrderedDict
from excel2xx.fieldmeta import FieldMeta
from excel2xx import console, utils


def toKV(sheet, fields=("key", "type", "value")):
    d = OrderedDict()
    rowNum = 0
    for row in sheet:
        rowNum += 1
        key = row[fields[0]]
        valType = row[fields[1]]
        val = row[fields[2]]
        if isinstance(val, float):
            val = intIf(val)
            pass
        meta = FieldMeta.parseField(key, valType)
        if key in d:
            raise Exception(f"duplicated key: {key}")
        try:
            d[key] = meta.format(val)
        except Exception as ex:
            console.error(
                "Field(name=%s type=%s). row: %s col: %s\n\t err: %s\n\t %s\n"
                % (
                    meta.name,
                    meta.type,
                    utils.show_row(rowNum),
                    utils.show_col(3),
                    str(ex),
                    val,
                )
            )
            pass
        pass
    return d


def intIf(val):
    if isinstance(val, float):
        if val == float(int(val)):
            return int(val)
    return val
